#include <stdatomic.h>
#include <inttypes.h>
#include "deque.h"
#include <stdlib.h>
#include <stdio.h>

Deque * create() {
	Deque * q = (Deque *) calloc(1, sizeof(Deque));
	Array * a = (Array *) calloc(1, sizeof(Array)+2*sizeof(atomic_int));
	atomic_store_explicit(&q->array, a, memory_order_relaxed);
	atomic_store_explicit(&q->top, 0, memory_order_relaxed);
	atomic_store_explicit(&q->bottom, 0, memory_order_relaxed);
	atomic_store_explicit(&a->size, 2, memory_order_relaxed);
	return q;
}


/**
	@Begin
	@Interface_define: Take 
	@End
*/
int take(Deque *q) {
	size_t b = atomic_load_explicit(&q->bottom, memory_order_relaxed) - 1;
	/**
		@Begin
		@Commit_point_define_check: true
		@Label: Take_Read_Bottom
		@End
	*/
	Array *a = (Array *) atomic_load_explicit(&q->array, memory_order_relaxed);
	/**
		//@Begin
		@Commit_point_define_check: true
		@Label: Take_Read_Array
		@End
	*/
	atomic_store_explicit(&q->bottom, b, memory_order_relaxed);
	/**** SPEC (sequential) (testcase1.c) ****/
	atomic_thread_fence(memory_order_seq_cst);
	size_t t = atomic_load_explicit(&q->top, memory_order_relaxed);
	/**
		//@Begin
		@Commit_point_define_check: t != b
		@Label: Take_Read_Top
		@End
	*/
	int x;
	if (t <= b) {
		/* Non-empty queue. */
		int size = atomic_load_explicit(&a->size,memory_order_relaxed);
		x = atomic_load_explicit(&a->buffer[b % size], memory_order_relaxed);
		/**
			//@Begin
			@Commit_point_define_check: t != b
			@Label: Take_Read_Buffer
			@End
		*/
		if (t == b) {
			/* Single last element in queue. */
			//FIXME: weaken the following seq_cst causes no spec problem
			bool succ = atomic_compare_exchange_strong_explicit(&q->top, &t, t +
				1, memory_order_seq_cst, memory_order_relaxed);
			/**
				//@Begin
				@Commit_point_define_check: succ 
				@Label: Take_CAS_Top
				@End
			*/

			/**
				@Begin
				@Additional_ordering_point_define_check: true
				@Label: Take_Additional_Point
				@End
			*/
			if (!succ) {
				/* Failed race. */
				x = EMPTY;
			}
			atomic_store_explicit(&q->bottom, b + 1, memory_order_relaxed);
		}
	} else { /* Empty queue. */
		x = EMPTY;
		atomic_store_explicit(&q->bottom, b + 1, memory_order_relaxed);
	}
	return x;
}

void resize(Deque *q) {
	Array *a = (Array *) atomic_load_explicit(&q->array, memory_order_relaxed);
	size_t size=atomic_load_explicit(&a->size, memory_order_relaxed);
	size_t new_size=size << 1;
	Array *new_a = (Array *) calloc(1, new_size * sizeof(atomic_int) + sizeof(Array));
	size_t top=atomic_load_explicit(&q->top, memory_order_relaxed);
	size_t bottom=atomic_load_explicit(&q->bottom, memory_order_relaxed);
	atomic_store_explicit(&new_a->size, new_size, memory_order_relaxed);
	size_t i;
	for(i=top; i < bottom; i++) {
		atomic_store_explicit(&new_a->buffer[i % new_size], atomic_load_explicit(&a->buffer[i % size], memory_order_relaxed), memory_order_relaxed);
	}
	/**** detected UL ****/
	atomic_store_explicit(&q->array, new_a, memory_order_release);
	//printf("resize\n");
}

/**
	@Begin
	@Interface_define: Push
	@End
*/
void push(Deque *q, int x) {
	size_t b = atomic_load_explicit(&q->bottom, memory_order_relaxed);
	/**** SPEC (sequential) ****/
	size_t t = atomic_load_explicit(&q->top, memory_order_acquire);
	Array *a = (Array *) atomic_load_explicit(&q->array, memory_order_relaxed);
	if (b - t > atomic_load_explicit(&a->size, memory_order_relaxed) - 1) /* Full queue. */ {
		resize(q);
		// CDSSpec can actually detect the same bug if we avoid the UL error
		//Bug in paper...should have next line...
		//a = (Array *) atomic_load_explicit(&q->array, memory_order_relaxed);
		/**
			//@Begin
			@Commit_point_define_check: true
			@Label: Push_Read_Array
			@End
		*/
	}
	int size = atomic_load_explicit(&a->size, memory_order_relaxed);

	atomic_store_explicit(&a->buffer[b % size], x, memory_order_relaxed);
	/**
		//@Begin
		@Commit_point_define_check: true
		@Label: Push_Update_Buffer
		@End
	*/
	/**** SPEC (Sync) (run with -u100 to avoid the uninitialized bug) ****/
	atomic_thread_fence(memory_order_release);
	// Watch out!!! We need to specifiy the fence as the commit point because
	// the synchronization in C11 actually is established at the fence.
	/**
		@Begin
		@Commit_point_define_check: true
		@Label: Push_Update_Bottom
		@End
	*/
	atomic_store_explicit(&q->bottom, b + 1, memory_order_relaxed);
	
}

/**
	@Begin
	@Interface_define: Steal 
	@End
*/
int steal(Deque *q) {
	//Watch out: actually on need to be an acquire (don't count it)
	// An old bug
	size_t t = atomic_load_explicit(&q->top, memory_order_acquire);
	/**
		//@Begin
		@Potential_commit_point_define: true
		@Label: Steal_Potential_Read_Tail
		@End
	*/
	//FIXME: remove the fence causes no error and fewer executions..
	atomic_thread_fence(memory_order_seq_cst);
	/**** SPEC & UL ****/
	size_t b = atomic_load_explicit(&q->bottom, memory_order_acquire);
	/**
		@Begin
		@Commit_point_define_check: true 
		@Label: Steal_Read_Bottom
		@End
	*/

	/**
		//@Begin
		@Commit_point_define: t >= b
		@Potential_commit_point_label: Steal_Potential_Read_Tail
		@Label: Steal_Read_Tail
		@End
	*/
	int x = EMPTY;
	if (t < b) {
		/* Non-empty queue. */
		/**** detected UL ****/
		Array *a = (Array *) atomic_load_explicit(&q->array, memory_order_acquire);
		/**
			//@Begin
			@Commit_point_define_check: true
			@Label: Steal_Read_Array
			@End
		*/
		int size = atomic_load_explicit(&a->size, memory_order_relaxed);
		x = atomic_load_explicit(&a->buffer[t % size], memory_order_relaxed);
		/**
			//@Begin
			@Potential_commit_point_define: true
			@Label: Steal_Potential_Read_Buffer
			@End
		*/
		/**** SPEC (sequential) ****/ 
		bool succ = atomic_compare_exchange_strong_explicit(&q->top, &t, t + 1,
			memory_order_seq_cst, memory_order_relaxed);
		/**
			//@Begin
			@Commit_point_define_check: succ
			@Label: Steal_CAS_Top
			@End
		*/

		/**
			@Begin
			@Additional_ordering_point_define_check: true 
			@Label: Steal_Additional_Point
			@End
		*/

		/**
			//@Begin
			@Commit_point_define: succ
			@Potential_commit_point_label: Steal_Potential_Read_Buffer
			@Label: Steal_Read_Buffer
			@End
		*/
		if (!succ) {
			/* Failed race. */
			return ABORT;
		}
	}
	return x;
}
